#!/usr/bin/env python
# -*- python -*-
#
#      Copyright (C) 2002-2004 Robert Muth <robert at muth dot org>
#
#      This program is free software; you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation; version 2 of June 1991.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program in form of the file COPYING; 
#      if not, write to the 
#
#      Free Software Foundation, Inc. <http://www.fsf.org>
#      59 Temple Place, Suite 330, 
#      Boston, MA 02111-1307  USA
#
######################################################

__version__ = "Cjukebox $Revision: 1.37 $"

#######################################################
## CONFIG
#######################################################

## This where all the playlists are
## including implicit ones like good.m3u and bad.m3u

ConfigPlaylistRoot = "/home/music/0_Playlist"

## starting directory for directory browsing windows

ConfigMusicRoot = "/home/music"

## log file for messages when status panel is not available

ConfigLogFile   = "cjukebox.log"

## scritpt invoked to show stuff to the user

ConfigShowInfoScript = "./cjukebox.info.sh"

## line used in help screen

ConfigLine = "=" * 70

## line used in help screen

ConfigUseLogfileAndStatusWindow = 1

#######################################################
##  STANDARD IMPORTS
#######################################################

import os
import sys
import time
import getopt
import signal
import string
#import select
#import re
import curses
import curses.textpad
import whrandom
from curses import panel

#######################################################
## LOCAL IMPORTS
#######################################################


import musicus
#import playlist
import song 
#import cddb

#######################################################
## Debug Helpers
#######################################################

def ObjectInfo(obj):
    s = ""
    for x in dir(obj):
        s += x + "\n"
        s += "\t" +  str(getattr(obj,x).__doc__) + "\n" # __doc__ maybe None
    return s

#######################################################
##
#######################################################

def SaveToFile(file, message):
    file = open(file,"wa")
    file.write( message )
    file.close()
    return

#######################################################
## 
#######################################################

def StringMinSec(secs):
    return "%d:%02d" % divmod( secs,60)
#######################################################
## 
#######################################################


logfile = None
GlobalDisplayManager = None

def Message(prefix,message):
    global logfile

    if not GlobalDisplayManager or ConfigUseLogfileAndStatusWindow:
        if not logfile: logfile = open(ConfigLogFile,"wa")
        prefix  = prefix + " " + time.strftime("%Y/%m/%d %H:%M")
        logfile.write(prefix + ": " + message + "\n")

    if GlobalDisplayManager:
        panel = GlobalDisplayManager.status_panel()
        pos = panel.line_append(message,1)
        panel.pos_cursor_set(pos,1) # move to end of list

    return


def Debug(message):
    Message("D",message)
    return

def Info(message):
    Message("I",message)
    return

def Warning(message):
    Message("W",message)
    return

def Error(message):
    Message("E","ERROR: " + message)
    return
    
def Fatal(message):
    Message("F",message)
    sys.exit(-1)
    return

#######################################################
## 
#######################################################

class Keymap:

    GLOBAL_KEY = 1
    def __init__(self):
        self.methods = [None] * curses.KEY_MAX

    def bind(self, key, method, args, desc):
        self.methods[key] = (method, args, desc, 1)

    def gbind(self, key, method, args, desc):
        self.methods[key] = (method, args, desc, 0)

    def is_bound(self,key):
        if self.methods[key] == None:
            return 0
        return 1

    def process(self, key, focus):
        if self.is_bound(key):
            method, args, desc, f  = self.methods[key]
            if focus != 0 or f == 0:
                apply(method, args)
                return 1
        return 0

    def help_strings(self):
        message = []
        for k in range(curses.KEY_MAX):
            if self.methods[k] == None:
                continue
            method, args, desc, focus  = self.methods[k]
            if desc:
                if focus:
                    line = "* "
                else:
                    line = "  "
                k = desc[0]
                k += " " * 30
                line += k[0:25] + desc[1] 
                message.append(line)
        return message

#######################################################
## 
#######################################################

class TimerEvents:
    def __init__(self):
        self.tag = 0
        self.events = {}

    def add_onetime(self, timeout, func, args):
        self.tag = self.tag + 1
        self.events[self.tag] = (func, args, time.time() + timeout, -1)
        return self.tag

    def add_periodic(self, period, func, args):
        self.tag = self.tag + 1
        self.events[self.tag] = (func, args, time.time() + period, period)
        return self.tag

    def remove(self, tag):
        if self.events.has_key(tag):
            del self.events[tag]
        return
    
    def poll(self):
        now = time.time()

        for tag, (func, args, timeout, period) in self.events.items():
            if now >= timeout:
                apply(func, args)
                if period < 0.0:
                    self.remove(tag)
                else:
                    self.events[tag] = (func, args, now + period, period)
        return
    
#######################################################
## 
#######################################################


#######################################################
## 
#######################################################

class ModalDialog:
    def __init__(self,x, y, w, h):
        self._window = curses.newwin(h, w, y, x)
        self._textbox = curses.textpad.Textbox(  self._window )
        self._textbox.edit()
        return
    
#######################################################
## Generic Panel Class 
#######################################################

class Panel:

    def __init__(self,parent,name, x, y, w, h):
        self._window = curses.newwin(h, w, y, x)

        self._pan = panel.new_panel(self._window)
        self._dim = (x,y,w,h)
        self._screen_h = h - 2
        self._screen_w = w - 2

        self._parent = parent
        self._children = []

        if parent:
            parent.child_add(self)

        self._name = name
        self._keymap = Keymap()
        self.focus_off()
        self.visible_off()
        return

    FLAG_NORMAL   = 0
    FLAG_CURSOR   = 1
    FLAG_PLAYING  = 2    
    FLAG_MARKED   = 4
    FLAG_TITLE    = 8

    def resize(self,x,y,w,h):
        self._window.resize(h, w) # reverse args, it seems to be important to call resize() before move()
        self._pan.move(y,x)       # reverse args
        self._dim = (x,y,w,h)
        self._screen_h = h - 2
        self._screen_w = w - 2
        return
    
    def draw_set_attributes(self,flags):
        self._window.attroff(curses.A_BLINK)
        self._window.attroff(curses.A_REVERSE)
        self._window.attroff(curses.A_BOLD)
        
        if flags & Panel.FLAG_PLAYING:
            self._window.attron(curses.A_BLINK)
            
        if flags & (Panel.FLAG_CURSOR|Panel.FLAG_MARKED):
            self._window.attron(curses.A_REVERSE)

        if flags & (Panel.FLAG_TITLE|Panel.FLAG_CURSOR):
            self._window.attron(curses.A_BOLD)

        return 


    def draw_clear(self):
        self.draw_set_attributes(Panel.FLAG_NORMAL)
        self._window.clear()
        return
        
    def draw_header(self,pos,string, flags= FLAG_TITLE):
        self.draw_set_attributes(flags)
        if pos < self._screen_w:
            self._window.addnstr(0,pos,string, self._screen_w - pos)
        return 

    def draw_footer(self,pos,string,flags= FLAG_TITLE):
        self.draw_set_attributes(flags)
        if pos < self._screen_w:
            self._window.addnstr(self._screen_h+1,pos,string,self._screen_w - pos)  
        return
    
    def draw_box(self):
        self.draw_set_attributes(Panel.FLAG_NORMAL)
        self._window.box()    
        return
    
    def redraw(self,justborder = 0):
        if not  self._visible:
            Info("window  %s" % self._name )
        assert( self._visible )
        if not justborder:
            self.draw_clear()
            self.redraw_center() # "virtual function" 
        self.draw_box()
        self.redraw_border()   # "virtual function"
        return
    
    def draw_row(self,pos,str,flags):
        if pos < 0 or pos >= self._screen_h:
            Warning("out of bounds line drawing\n")

        self.draw_set_attributes(flags)
           
        self._window.move(pos+1,1)

        str += ' ' * self._screen_w
        str = str[0:self._screen_w]

        self._window.addstr(str)        
        return

    def draw_frame(self,flags=0):
        self.draw_set_attributes(flags)
#         self._window.border('*','*','*','*','*')
        self._window.border('|','|','-','-','+','+','+','+')
        return
            
    def child_add(self,child):
        self._children.append(child)
        return
    
    def visible_on(self):
        self._pan.show()
        self._pan.top()
        self._visible = 1
        self.redraw()
        return

    def is_visible(self):
        # should really be the opposite at all times
        assert( self._pan.hidden() != self._visible )
        return self._visible

    def visible_off(self):
        self._visible = 0
        self._pan.bottom()
        self._pan.hide()
        return

    def focus_on(self):
        self._focus = 1
        if curses.has_colors():
            self._window.attron( curses.color_pair(2) )
        else:
            self._window.attron(curses.A_BOLD)
        return
        
    def focus_off(self):
        self._focus = 0
        if curses.has_colors():
            self._window.attron( curses.color_pair(1) )
        else:
            self._window.attroff(curses.A_BOLD)
        return
        
    def has_focus(self):
        return self._focus
        
    def process_key(self, key):
        if self._keymap.process(key,self._focus):
            return 1
        for child in self._children:
            if child.process_key(key):
                return 1
        
    def string_status(self):
        return "panel %s:  v:%d f:%d" % (self._name ,self._visible,self._focus)

    def help_strings(self):

        message = []
 
        message.append(ConfigLine)
        message.append(self._name)
        message.append(ConfigLine)
        
        message += self._keymap.help_strings()
        
        return message

    
#######################################################
## fake panel covering the entire screen
## has some keyboard events attached to it        
#######################################################

class PanelRoot(Panel):

    def __init__(self, parent, name, x, y , w, h):
        Panel.__init__(self, parent, name, x, y, w, h)

        #self._keymap.bind(12, self.update, ()) # C-l
        self._keymap.gbind(ord('q'), self.quit, (), ("f10,q","Quit"))
        self._keymap.gbind(curses.KEY_F10, self.quit, (), None)

        self._keymap.gbind(ord('h'), self.toggle_help, (), ("f1,h","Toggle Help Screen"))
        self._keymap.gbind(curses.KEY_F1, self.toggle_help, (), None)


        self._keymap.gbind(curses.KEY_SLEFT, self.move_focus_left, (), ("shift <-/->","Move focus to left/right window"))
        self._keymap.gbind(curses.KEY_SRIGHT, self.move_focus_right, (), None)
        self._keymap.gbind(ord('\t'), self.cycle_focus, (), ("tab","Cycle focus through all windows"))
#        self._keymap.gbind(curses.KEY_RIGHT, self.cycle_next, (), "c:\tNext Screens")

        self._keymap.gbind(ord('('), self.resize_horizontal, (-1,),("(,)","Move vertical window split left/right"))
        self._keymap.gbind(ord(')'), self.resize_horizontal, (+1,),None)

        self._pan.bottom()
        self._pan.hide()
        return

    def quit(self):
        GlobalDisplayManager.quit()
        return
    
    # cycle focus through all children


    def cycle_focus(self):
        GlobalDisplayManager.cycle_focus()
        return 

    def move_focus_right(self):
        GlobalDisplayManager.move_focus_right()
        return 
    
    def move_focus_left(self):
        GlobalDisplayManager.move_focus_left()
        return 
    
    def resize_horizontal(self,delta):
        GlobalDisplayManager.resize_horizontal(delta)
        return
        
    
    def toggle_help(self):
#        self.redraw()

        if self.is_visible():
            self.visible_off()
        else:
            self.visible_on()
        return
    
    def redraw_border(self):
       self.draw_frame()
       self.draw_header(1,"[" +  __version__ + "]")
       self.draw_footer(1,"[Help - Press 'h' to exit]")
       return
   
    def redraw_center(self):
        i=0
        self.draw_row(i,"",Panel.FLAG_NORMAL)
        i+=1
        if 0:
            self.draw_row(i,ConfigLine)
            i += 1
            self.draw_row(i,"Players")
            i += 1
            self.draw_row(i,ConfigLine)
            i += 1

            messages = GlobalPlayers.help_strings()
            for l in messages:
                self.draw_row(i,l)
                i += 1

        messages = GlobalDisplayManager.help_strings()
        for l in messages:
            self.draw_row(i,l,Panel.FLAG_NORMAL)
            i += 1
        while i < self._screen_h:
            self.draw_row(i,"",Panel.FLAG_NORMAL)
            i+=1
        return
    
#######################################################
## Scrollable Panel Class
#######################################################

class ScrollablePanel(Panel):
    """ manages lines and cursor position
    there can be more lines than fit on the screen
    """

    NOREDRAW = 1
    
    def __init__(self,parent,name, x, y, h, w,allow_marking):
        allow_marking = 0 # disable this for now

        Panel.__init__(self, parent, name, x, y, h, w)
        self._flags = []
        self._rows = []
        self._cursor_pos = 0
        self._screen_pos = 0

        self._keymap.bind(curses.KEY_UP, self.pos_cursor_move, (-1,1),("up/dn","Scroll up/down")) 
        self._keymap.bind(curses.KEY_DOWN, self.pos_cursor_move, (1,1),None) 

        self._keymap.bind(curses.KEY_PPAGE, self.pos_cursor_move, (-h+5,1),("pgup/pgdn","Scroll page up/down"))
        self._keymap.bind(curses.KEY_NPAGE, self.pos_cursor_move, (h-5,1),None)

        self._keymap.bind(curses.KEY_HOME, self.pos_cursor_move, (-100*1000,1),("home/end","Scroll home/end"))
        self._keymap.bind(curses.KEY_END, self.pos_cursor_move, (100*1000,1),None)

        if allow_marking:
            self._keymap.bind( ord('m'), self.current_pos_mark, (),("m/M","mark/unmark"))
            self._keymap.bind( ord('M'), self.current_pos_unmark, (),None)

            self._keymap.bind( ord('a'), self.mark_all, (),("a/A","mark/unmark all"))
            self._keymap.bind( ord('A'), self.unmark_all, (),None)
        return

    def pos_is_visible(self,pos):
        if pos < self._screen_pos: return 0
        if pos >= self._screen_pos + self._screen_h: return 0
        return 1

    def pos_make_visible(self, pos):
        assert( pos >= 0 )
        
        if pos < self._screen_pos:
            self._screen_pos = pos
            self.redraw_lines()

        elif pos >= self._screen_pos+self._screen_h:
            self._screen_pos = pos - self._screen_h +1
            if self._screen_pos < 0:
                self._screen_pos = 0
            self.redraw_lines()
        return


    def find_pos_with_flags(self,flags):
        for pos in range(len(self._flags)):
            if self._flags[pos] & flags == flags: return pos
        return None

    def line_change_flags(self, pos, flags, redraw):
        self._flags[pos] = flags
        if redraw:
            self.redraw_line( pos )
        return
    
    def line_add_flags(self, pos, flags, redraw):
#        assert( flags &  Panel.FLAG_CURSOR == 0)
        self.line_change_flags(pos, self._flags[pos] | flags,redraw)
        return
    
    def line_sub_flags(self, pos, flags, redraw):
#        assert( flags &  Panel.FLAG_CURSOR == 0)
        self.line_change_flags(pos, self._flags[pos] & ~flags,redraw)
        return
        
    def pos_cursor_set(self, newpos, redraw):
        if len(self._rows) <= 0: return

        assert( self._flags[self._cursor_pos] & Panel.FLAG_CURSOR )

        self.line_sub_flags(self._cursor_pos, Panel.FLAG_CURSOR, redraw )

        if newpos < 0: newpos = 0
        if newpos >= len(self._rows): newpos = len(self._rows) - 1
        # update cursor pos
        self._cursor_pos = newpos

        assert( self._cursor_pos >= 0 )

        self.pos_make_visible( self._cursor_pos )

        self.line_add_flags(self._cursor_pos, Panel.FLAG_CURSOR, redraw)
        assert( self._flags[self._cursor_pos] & Panel.FLAG_CURSOR )
        return

    def pos_cursor_move(self,delta,redraw):
        self.pos_cursor_set(self._cursor_pos + delta, redraw )
        return

    def pos_cursor_get(self):
        assert( self._flags[self._cursor_pos] & Panel.FLAG_CURSOR )
        return self._cursor_pos

    def pos_random_get(self):
        return whrandom.randint(0, len(self._rows) - 1)
    
        
    def redraw_line(self, pos):
        if not self.pos_is_visible(pos): return
        
        if pos < len(self._rows):
            flags = self._flags[pos] 
            str = self._rows[pos]
        else:
            flags = Panel.FLAG_NORMAL
            str = ""

        self.draw_row(pos-self._screen_pos ,str,  flags)
        return
                    
    def redraw_lines(self):
        for pos in range(self._screen_h):
            self.redraw_line(self._screen_pos + pos)
        return
    
    def lines_clear(self,redraw=1):
        self._rows = []
        self._flags = []
        self._cursor_pos = 0
        self._screen_pos = 0
        if redraw:
            self.redraw_lines()
        return

    def line_delete(self, pos, redraw):
        if self._cursor_pos == pos:
            self.pos_cursor_move(1,redraw)

        if self._cursor_pos == pos:
            self.pos_cursor_move(-1,redraw)
            
        del self._flags[pos]
        del self._rows[pos]
        
        if self._cursor_pos >= pos:
            self._cursor_pos -= 1
        if redraw:
            self.redraw_lines()
        return

    def line_insert(self, pos, str, redraw):

        old_num_rows = len(self._rows)
        self._flags.insert(pos, Panel.FLAG_NORMAL)
        self._rows.insert(pos, str)

        if self._cursor_pos >= pos:
            self._cursor_pos += 1
        if old_num_rows <= 0:
            self._cursor_pos = 0
            self._screen_pos = 0
            self._flags[0] |= Panel.FLAG_CURSOR            
        if redraw:
            self.pos_make_visible( self._cursor_pos )
        #
        # this is an important speed optimization:
        # for complete directory changes
        #
            if pos == old_num_rows and pos != self._cursor_pos: 
                self.redraw_line(pos)
            else:
                self.redraw_lines()
        return

    def line_swap(self, pos1, pos2, redraw):

        l1 = self.line_get(pos1)
        l2 = self.line_get(pos2)

        self.line_set( pos1, l2, redraw)
        self.line_set( pos2, l1, redraw)

        if l1[0] &  Panel.FLAG_CURSOR: self._cursor_pos = pos2
        elif l2[0] &  Panel.FLAG_CURSOR: self._cursor_pos = pos1
        
        self.pos_make_visible( self._cursor_pos )
        return
    
    def line_append(self,str,redraw):
        pos = len(self._rows)
        self.line_insert( pos ,str,redraw)
        return pos
    
    def line_set(self, pos, (flags,str), redraw):
        self._rows[pos] = str
        self._flags[pos] = flags
        if redraw:
            self.redraw_line(pos)
        return

    def line_get(self,pos):
        return (self._flags[pos],self._rows[pos]) 

    def mark_all(self):
        for pos in range(len(self._flags)):
            self.line_add_flags( pos,Panel.FLAG_MARKED,1)
        return

    def unmark_all(self):
        for pos in range(len(self._flags)):
            self.line_sub_flags( pos,Panel.FLAG_MARKED,1 )
        return
    
    def current_pos_mark(self):
        self.line_add_flags( self._cursor_pos,Panel.FLAG_MARKED,1)
        self.pos_cursor_move(1,1)
        return

    def current_pos_unmark(self):
        self.line_sub_flags( self._cursor_pos,Panel.FLAG_MARKED,1)
        self.pos_cursor_move(1,1)
        return
    
                                          
    def redraw_center(self):
        self.redraw_lines()
        return
#######################################################
## PLAYBACK WINDOW
#######################################################

class PanelStatus(ScrollablePanel):

    def __init__(self,parent,name, x, y, h, w):
        ScrollablePanel.__init__(self, parent, name, x, y, h, w,0)
        return

    def redraw_border(self):   #virtual function
        self.draw_frame()
        self.draw_header(2,"[STATUS]")
        return

#######################################################
## 
#######################################################

class PanelList(ScrollablePanel):
    """ this window class knows about music files and how to render them"""
    def __init__(self, parent, name, x, y, h, w,allow_marking):
        ScrollablePanel.__init__(self, parent, name, x, y, h, w,allow_marking)
        self._playlist = song.Playlist()  
        self._viewmode = song.Viewmode()
   
        self._keymap.bind(ord('v'), self.viev_mode_cycle, (),("v","Cycle view mode"))
        return

    def viev_mode_cycle(self):
        self._viewmode.cycle()

        for i in range(self._playlist.length()):
            flags = self.line_get(i)[0]
            item = self.item_get(i)
            str = item.string_fancy_get(self._viewmode)
            self.line_set(i, (flags,str),ScrollablePanel.NOREDRAW)
        self.redraw()
        return
    
    def items_clear(self,redraw=1):
        self.lines_clear(redraw)
        self._playlist.clear("*noname*")
        return

    def item_append(self,item, redraw=1):
        str = item.string_fancy_get(self._viewmode)
        self._playlist.append(item)
        return self.line_append(str,redraw)

    def item_delete(self, pos, redraw=1):
        self.line_delete(pos, redraw)
        self._playlist.delete(pos)
        return

    def item_swap(self, pos1, pos2, redraw=1):
        self._playlist.swap(pos1,pos2)
        self.line_swap(pos1,pos2,redraw)
        return
    
    def item_insert(self,pos,item,redraw=1):
        str = item.string_fancy_get(self._viewmode)
        self.line_insert(pos,str,redraw)
        self._playlist.insert(pos,item)
        return

    def item_get(self,pos):
        return self._playlist.item( pos )
            
    def item_get_current(self):
        if len( self._rows) <= 0:
            return None
        return self._playlist.item( self.pos_cursor_get() )

    def items_update_from_playlist(self,redraw=1):
        for item in  self._playlist.items():
            str = item.string_fancy_get(self._viewmode)
            self.line_append(str,redraw)
        return

#######################################################
## PLAYBACK WINDOW
#######################################################

class PanelSong(PanelList):

    def __init__(self,parent,name, x, y, h, w):
        PanelList.__init__(self, parent, name, x, y, h, w,0)
        self._current_item = None
        self._callback_abort = None
        self._callback_finished = None
        self._oldpos = 0
        self._paused = 0
        self._skip_bad_songs = 1
        GlobalTimer.add_periodic(0,self.timer_based_polling,())
        self._time_bad_list = ""
        self._time_good_list = ""

        self._bad_list = song.Playlist()
        self._bad_list.load(ConfigPlaylistRoot + "/bad.m3u",GlobalMusicus)

        self._good_list = song.Playlist()
        self._good_list.load(ConfigPlaylistRoot + "/good.m3u",GlobalMusicus)
        self.init_keybindings()
        return
    
    def init_keybindings(self):

        ## all keybinding here a global, .i.e. they are active even if we do not have focus
        #        self._keymap.gbind(ord('-'), self.change_volume, (-3,),("-/+","Vol -/+"))
        #        self._keymap.gbind(ord('+'), self.change_volume, (+3,),None)
        self._keymap.gbind(ord('p'), self.toggle_pause,(),("p","Toggle pause"))
        self._keymap.gbind(ord('k'), self.kill_song,(),("k","Stop Song/Playlist"))
        self._keymap.gbind(ord('i'), self.show_info,(),("i","show info for current song"))
        self._keymap.gbind(ord(' '), self.skip_song,(),("space:","Skip Song"))
        self._keymap.gbind(ord('g'), self.add_to_good_list,(),("g/b","Add to Goodlist/Badlist"))
        self._keymap.gbind(ord('b'), self.add_to_bad_list,(),None)
        self._keymap.gbind(ord('s'), self.toggle_bad_song_skipping,(),("s","Toggle Bad Song Skipping"))

        
        self._keymap.gbind(ord('r'), self.song_move, (-2000,), ("<-,r","Fast Rewind"))
        self._keymap.gbind(curses.KEY_LEFT, self.song_move, (-2000,), None)

        self._keymap.gbind(ord('f'), self.song_move, (-2000,), ("->,f","Fast Forward"))
        self._keymap.gbind(curses.KEY_RIGHT, self.song_move, (2000,), None)

        return

    def toggle_bad_song_skipping(self):
        self._skip_bad_songs = not self._skip_bad_songs
        self.redraw()
        return
    
    def add_to_bad_list(self):
        if self._current_item == None: return
        if self._bad_list.contains( self._current_item ): return
        Info("adding to bad list: " + self._current_item.string_get() )

        timestr = time.strftime("%Y/%m/%d %a")
        if self._time_bad_list != timestr:
            self._time_bad_list = timestr
            self._bad_list.append( song.Item("# " + timestr ,GlobalMusicus) )

        self._bad_list.append( self._current_item )
        self._bad_list.save()
        self.skip_song()
        return

    def add_to_good_list(self):
        if self._current_item == None: return
        if self._good_list.contains( self._current_item ): return

        Info("adding to good list: " + self._current_item.string_get() )
        timestr = time.strftime("%Y/%m/%d %a")
        if self._time_good_list != timestr:
            self._time_good_list = timestr
            self._good_list.append( song.Item("# " + timestr, GlobalMusicus) )
            
        self._good_list.append( self._current_item )
        self._good_list.save()
        return

    def show_info(self):
        if self._current_item == None: return
        cmd = ConfigShowInfoScript + " " + self._current_item.description_get(self._viewmode) + "&>/dev/null&"
        Info("Invoking info command: " + cmd )
        os.system( cmd )
        return

    def kill_song(self):
        if self._current_item:
            self._current_item = None
#            Info("Killing current song")
            if GlobalMusicus: GlobalMusicus.stop()
            if self._callback_abort:
                self._callback_abort()
        self.redraw()
        return
    
    def skip_song(self):
        if self._current_item:
            self._current_item = None
            Info("Skipping current song")
            if GlobalMusicus: GlobalMusicus.stop()
            if self._callback_finished:
                self._callback_finished()
        else:
            Info("No song active")
        self.redraw()
        return
    
    def timer_based_polling(self):

        if self._current_item == None: return
        #
        # there appears to be a bug the plugins which
        # may result in negative positions to be returned
        # we fix this by getting a second opinion after some delay 
        #
        pos = self.get_pos()
        if pos < 0:
            self.skip_song()
        if pos != self._oldpos:
            self._oldpos = pos
            self.redraw(1)
        return
    
    def get_pos(self):

        # need to test GlobalMusicus since no song would be playing without it
        result = GlobalMusicus.getpos()
        if result[0] == 0:
            pos = int(result[1])
            #Info("got pos %d\n" %pos)
            return pos
        else:
            return -1

    def get_length(self):
        # need to test GlobalMusicus since no song would be playing without it
        result = GlobalMusicus.getlength()
        if result[0] == 0:
            return int(result[1])
        else:
            return -1

    def song_move(self,msec):
        # need to test GlobalMusicus since no song would be playing without it
        pos = self.get_pos()
        if pos < 0: return
        pos = pos + msec
        if pos < 1: pos = 1
        GlobalMusicus.setpos(pos)
        return
    
    def play_song(self,item,callback_abort,callback_finished):
        self._paused = 0
        if not item.is_song():
            Warning("Item is not a song " + file )
            return 0

        file =  item.string_get()
        if not os.access(file,os.R_OK):
            Warning("File not readable " + file )
            return 0
        
        if self._skip_bad_songs and self._bad_list.contains(item):
            Info("Never list match, skipping" + file )
            return 0

        # if we are already playing a song kill it

        self.kill_song()
        Info("Playing " + file )
        if GlobalMusicus: result = GlobalMusicus.play(file)
        if result[0]:
            Error("Player not found - unknown codec for " + file )
            return 0

        #
        # this delay loop is necessary because???
        # The mpc plug in appears to be buggy 
        #
        
        time.sleep(0.05)

        self.get_length()
        
        while self.get_pos() < 0:
            pass
        
        self._callback_abort = callback_abort
        self._callback_finished = callback_finished
        pos = self.item_append(item,ScrollablePanel.NOREDRAW)
        self.pos_cursor_set(pos,1) # move to end of list
        
        #self.redraw()
        # !!! this one must be delayed as long as possible since
        # it is the semaphore for the timer based polling
        self._current_item = item
                
        return 1

    def get_volume(self):
        # TODO: check GlobalMusicus
        result = GlobalMusicus.getvol()
        if result[0] == 0:
            if( int(result[1]) > 100 ): return 100
            if( int(result[2]) > 100 ): return 100
            assert( int(result[1]) <=100)
            assert( int(result[2]) <=100)
            return (int(result[1]) + int(result[2]))/2
        else:
            return 0

    def change_volume(self,delta):
        # TODO: check GlobalMusicus
        vol = self.get_volume() + delta
        result = GlobalMusicus.setvol(vol,vol)
        Info("changing volume to %d\n" % vol )
        self.redraw(1)
        return
    
    def toggle_pause(self):
        if self._current_item:
            if self._paused:
                self._paused = 0
                GlobalMusicus.unpause()
            else:
                self._paused = 1
                GlobalMusicus.pause()
            self.redraw_border()    
        return

    def redraw_border(self):   #virtual function
        self.draw_frame()
        self.draw_header(2,"[PLAY-HISTORY]")
        self.draw_footer(2,"[" + self._viewmode.string() + "]")

        self.draw_footer(15,"[badlist %3d]" % self._bad_list.length())
        self.draw_footer(30,"[goodlist %3d]" % self._good_list.length())

#        self.draw_footer(40,"[Vol %3d]" % self.get_volume())

        if self._skip_bad_songs:
            self.draw_footer(50,"[skip]")

        if self._current_item and self._paused:
            self.draw_footer(60,"[paused]",  Panel.FLAG_TITLE | Panel.FLAG_PLAYING)

        if self._current_item:
            self.draw_footer(70,"[" + StringMinSec( self.get_pos()/1000) + "/" +
                             StringMinSec(self.get_length()/1000) + "]")
        

            #self.draw_footer(10,"%d" % self._song._pid)
            #self.draw_footer(20,"(%d,%d)" %q self._song._result)
        return
    
        
#######################################################
## PLAYLIST EDITOR WINDOW
#######################################################


class PanelListPlaylistEditor(PanelList):
    """ editor for playlist"""
    global ConfigPlaylistRoot

    def __init__(self,parent,name, x, y, h, w):
        PanelList.__init__(self, parent, name, x, y, h, w,1)

        self._mode_playback = ["normal","shuffle","repeat"] 
        self._current_play_pos = -1
        self.init_keybindings()
        
        return

    def init_keybindings(self):
        self._keymap.bind(ord('\n'), self.play_file, (),("enter","Play file"))
        #self._keymap.bind(curses.KEY_ENTER, self.play_file, (),"Play file")

#        self._keymap.bind(ord('n'), self.new_play_list, (),("c","New Playlist"))
        
        self._keymap.bind(ord('u'), self.song_move_current, (-1,),("u/d","Move line up/down")) 
        self._keymap.bind(ord('d'), self.song_move_current, (1,),None)

#        self._keymap.bind( curses.KEY_SDOWN, self.song_move_current, (1,),None)
#        self._keymap.bind( curses.KEY_SUP, self.song_move_current, (-1,),None)
        
        self._keymap.bind(curses.KEY_F7, self.song_insert_current_comment, (),("f7","Insert comment"))

        self._keymap.bind(curses.KEY_F9, self.clear_dirty_flag, (),("f9","Force clearing of dirty flag"))
        self._keymap.bind(curses.KEY_BACKSPACE, self.exit_playlist, (),("backspace","Back to file browser "))
        
        self._keymap.bind(curses.KEY_F8, self.song_delete_current, (),("f8","Delete line"))
#        self._keymap.bind(curses.KEY_DELETE , self.song_delete_current, (),None)

        self._keymap.bind(curses.KEY_F2, self.save_current_filename, (),("f2","Save playlist"))

        #self._keymap.bind(curses.KEY_F4, self.del_line_current, (),"f3:\tEdit Line",1)

        ##
        self._keymap.bind(ord('c'), self.toggle_playback_mode, (),("c","Cycle playback mode"))

#        self._keymap.bind(curses.KEY_F5, self.append_to_playlist, (0,0),("f5","Append to playlist"))
        self._keymap.bind(curses.KEY_F5, self.transfer_item, (1,0),("f5","Append to other window"))        
#        self._keymap.bind(curses.KEY_F6, self.append_to_playlist, (0,0),("f6","Insert into  playlist"))
        self._keymap.bind(curses.KEY_F6, self.transfer_item, (1,1),("f6","Insert into other window"))        
        return        

    def clear_dirty_flag(self):
        Warning("dirty flag was forced to clear")
        self._playlist.clear_dirty_flag()
        assert( not self._playlist.dirty() )
        self.redraw_border()
        return
    
    def exit_playlist(self):
        if self.dirty():
            Error("Dirty playlist must be saved first (press f9 to clear dirty bit)")
            return

        GlobalDisplayManager.exit_playlist_edit(self)
        return
    
    
    def toggle_playback_mode(self):
        self._mode_playback = self._mode_playback[1:] + self._mode_playback[0:1]
        Info("Changing Playmode")
        cmd = ConfigShowInfoScript + " new play mode is " +  self._mode_playback[0] + "&>/dev/nullh&"
        Debug("Info command: " + cmd )
        os.system( cmd )
        self.redraw()
        return

    def get_next_song(self,pos):
#        Info("get next starting %d" % pos)
        length = self._playlist.length()
        if length <= 0:
            return (0,None)

        if self._mode_playback[0] == "shuffle":
            for i in range(50):  
                pos = self.pos_random_get()
                obj = self.item_get(pos)
                if  obj.is_song():
                    return (pos,obj)
            return (0,None)
        
        else:
            for i in range(length):
                pos += 1
                if pos >= length:
                    if self._mode_playback[0] == "repeat":
                        Info("wrap around")
                             
                        pos = 0
                    else:
                        return (0,None)                
                obj = self.item_get(pos)
                if  obj.is_song():
                    return (pos,obj)
            return (0,None)    

    def clear_playing_flag(self):
        pos = self.find_pos_with_flags(Panel.FLAG_PLAYING)

        if pos != None:
            self.line_sub_flags(pos, Panel.FLAG_PLAYING,1 )
            
        assert( not self.find_pos_with_flags(Panel.FLAG_PLAYING) )
            
        return
    
    def play_finish(self):
        """ usually called when a song is finished
        """
        
        pos = self.find_pos_with_flags(Panel.FLAG_PLAYING)
        if not pos: pos = 0

        self.clear_playing_flag()  # get_next_song requires playing flag 
        
        for i in range(50):
            (pos,obj) = self.get_next_song(pos)

            if obj == None: break

            if GlobalDisplayManager.play_song(obj,self.play_abort,self.play_finish):
#                Info("setting pos %d" % pos)
                assert( not self.find_pos_with_flags(Panel.FLAG_PLAYING) )                
                self.line_add_flags(pos, Panel.FLAG_PLAYING,1 )
                return
        return
    
    def play_abort(self):
        self.clear_playing_flag()
        return
    
    def play_file(self):
        if self._playlist.length() <= 0:
            Error("Playlist is empty")
            return

        self.clear_playing_flag()  # get_next_song requires playing flag
        
        pos = self._cursor_pos
        obj = self.item_get(pos)
        if  obj.is_song() and GlobalDisplayManager.play_song(obj,self.play_abort,self.play_finish):
            assert( not self.find_pos_with_flags(Panel.FLAG_PLAYING) )
            self.line_add_flags(pos, Panel.FLAG_PLAYING,1 )
        else:
            Error("Cannot play: " + obj.string_get() )
        return

    def dirty(self):
        return self._playlist.dirty()
            
    def load_file(self,filename,is_new):
        if self.dirty():
            Error("Cannot overwrite dirty playlist")
            return

        self.items_clear(ScrollablePanel.NOREDRAW)
        if is_new:
            self._playlist.clear(filename)
        else:    
            self._playlist.load(filename,GlobalMusicus)
            self.items_update_from_playlist(ScrollablePanel.NOREDRAW)
        self.redraw()
        return
    
    def save_file(self,filename):
        Info("Saving playlist as " + filename)
        self._playlist.save(filename)
        self.redraw_border()
        return
    
    def save_current_filename(self):
        Info("Saving playlist")
        self._playlist.save()
        self.redraw_border()
        return
    
    def song_insert_current_comment(self):
        """ insert comment just before cursor """
        Info("Inserting comment" )

        item = song.Item("#",GlobalMusicus)

        if self._playlist.length() > 0:
            pos = self.pos_cursor_get()
            self.item_insert(pos,item,1)
        else:
            self.item_append(item)
            self.redraw_border()
        assert( self.dirty() )
        self.redraw_border()
        return
    
    def song_delete_current(self):
        """ delete line under cursor """
        if self._playlist.length() > 0:
            pos = self.pos_cursor_get()
            obj = self.item_get(pos)
            Info("Deleting song " + obj.string_get() )
                        
            self.item_delete(pos,ScrollablePanel.NOREDRAW)
            self.redraw() # FIXME there is a bug somewhere, so we use the big hammer for now
            #self.redraw_border()
            assert( self.dirty() )
        else:
            Error("Playlist is empty")
        return

    def song_move_current(self,delta):
        """ move song under cursor in given direction"""
        if self._playlist.length() > 0:
            pos1 = self.pos_cursor_get()
            pos2 = pos1 + delta
            if pos2 < 0: return
            if pos2 >= self._playlist.length(): return
            
            self.item_swap(pos1,pos2,1)
            self.redraw_border()
        else:
            Error("Playlist is empty")
        return

    def add_to_playlist_recursively(self, filename, at_cursor):
        """ recursive routine not called directly from gui"""
        obj = song.Item(filename,GlobalMusicus)
        if obj.is_song():
            if at_cursor and self._playlist.length() > 0:
                self.item_insert(self.pos_cursor_get(),obj,0)
            else:
                self.item_append(obj,ScrollablePanel.NOREDRAW)
            
        elif obj.is_dir():
            try:
                for file in os.listdir(filename):
                    pathname = os.path.join(filename, file)
                    self.add_to_playlist_recursively(pathname,at_cursor)
            except IOError:
                # errors are ok !!!
                pass
        return

    def add_item(self, item, at_cursor):
        """ add given song either at the end (at_cursor=0)
        or just before the cursur (at_cursor != 0)
        if song is a directory all song inside of it or any subdirectories will be added
        """
        if item.is_dir():
            self.add_to_playlist_recursively(item.string_get(), at_cursor)
        else:
            if at_cursor and self._playlist.length() > 0:
                self.item_insert(self.pos_cursor_get(),item,0)
            else:
                self.item_append(item,ScrollablePanel.NOREDRAW)
                
        self.redraw()
        return

    def transfer_item(self,recursive,at_cursor):
        obj =  self.item_get_current()
        if obj == None:
            Error("no item available")
        else:    
            GlobalDisplayManager.add_to_other_window(self,obj,at_cursor)
            self.pos_cursor_move(1,1)
            self.redraw()
        return

    def redraw_border(self):
        self.draw_frame()
        self.draw_header(2,"[PLAYLIST: " + self._playlist.filename() +"]")
        self.draw_footer(2,"[" + self._viewmode.string() + "]")

        self.draw_footer(15,"[" + self._mode_playback[0] + "]")

                
        self.draw_footer(30,"[" + StringMinSec( self._playlist.time() ) + "," + ("%d" % self._playlist.length()) + "]")

        if self.dirty(): self.draw_footer(50,"[dirty]")
        return

#######################################################
## FILE BROWSER WINDOW
#######################################################

def entry_compare(e1,e2):
    t1 = e1.type_get()
    t2 = e2.type_get()
    if t1 > t2: return 1
    elif t1 < t2: return -1
    
    p1 = e1.string_get()
    p2 = e2.string_get()
    
    if p1 > p2: return 1
    elif p1 < p2: return -1
    else: return 0
    
        
class PanelListFilebrowser(PanelList):
    """ browse directories similar to norton commander for DOS"""
    

    def __init__(self,parent,name, x, y, h, w):
        PanelList.__init__(self, parent, name, x, y, h, w,1)

        #self._keymap.bind(curses.KEY_ENTER, self.enter_button, (),"Change directory/Play song",1)

        self._playable_only = 1
        
        self.init_keybindings()
        return

    def init_keybindings(self):
        self._keymap.bind(ord('\n'), self.enter_button, (),("enter","Change directory/Play song/edit playlist"))

        self._keymap.bind(curses.KEY_BACKSPACE, self.change_dir_up, (),("backspace","Change directory up"))

        self._keymap.bind(curses.KEY_F3, self.new_playlist, (),("f3","New playlist"))

        self._keymap.bind(curses.KEY_F8, self.delete_file, (),("f8","Delete file"))

        self._keymap.bind(ord('t'), self.toggle_playable_only, (),("t","Toggle list playable (only show playable files)"))
#        self._keymap.bind(ord('r'), self.reload_toc, (),("r","Reload Toc"))

#        self._keymap.bind(curses.KEY_F5, self.append_to_playlist, (0,0),("f5","Append to playlist"))
        self._keymap.bind(curses.KEY_F5, self.transfer_item, (1,0),("f5","Append to other window"))        
#        self._keymap.bind(curses.KEY_F6, self.append_to_playlist, (0,0),("f6","Insert into  playlist"))
        self._keymap.bind(curses.KEY_F6, self.transfer_item, (1,1),("f6","Insert into other window"))
        
        self._keymap.bind(ord('0'), self.hotkey, (0,),("0,1,2,3,4,5,6,7,8,9","Load predefined playlist"))
        self._keymap.bind(ord('1'), self.hotkey, (1,),None)
        self._keymap.bind(ord('2'), self.hotkey, (2,),None)
        self._keymap.bind(ord('3'), self.hotkey, (3,),None)
        self._keymap.bind(ord('4'), self.hotkey, (4,),None)
        self._keymap.bind(ord('5'), self.hotkey, (5,),None)
        self._keymap.bind(ord('6'), self.hotkey, (6,),None)
        self._keymap.bind(ord('7'), self.hotkey, (7,),None)
        self._keymap.bind(ord('8'), self.hotkey, (8,),None)
        self._keymap.bind(ord('9'), self.hotkey, (9,),None)

        return
        

    def change_directory(self, dir, prefered="$$",redraw=1):
        
        Info("Change directory to " + dir ) # " prefered: " + prefered )
        
        if not os.access(dir,os.R_OK | os.X_OK): return

        dir = os.path.normpath(dir)
        self.items_clear(ScrollablePanel.NOREDRAW)
        self._cwd = dir
        itemlist = []

        #item = Item( os.path.join(dir,os.pardir) )
        #itemlist.append( item )

        for entry in os.listdir(dir):
            pathname = os.path.join(dir,entry)
            item = song.Item(pathname,GlobalMusicus)
            if self._playable_only and item.type_get() == song.Item.TYPE_FILE:
                continue
            itemlist.append( item )

        itemlist.sort( entry_compare )

        prefered_pos = None
        for entry in itemlist:
            pos = self.item_append(entry,ScrollablePanel.NOREDRAW)
            if prefered == entry.string_short_get():
                prefered_pos = pos 

        if prefered_pos:
            self.pos_cursor_set(prefered_pos,redraw)

        if redraw:
            self.redraw()
        return

    def reload_toc(self):
        song.Item.TocContainer.find_toc(self._cwd,0)
        self.redraw()
        return
            
    def change_dir_up(self):
        dummy, currdir = os.path.split(self._cwd)
        #assert (not currdir)
        newdir =  os.path.join(self._cwd,os.pardir)
        if not os.path.isdir(newdir): newdir = "/"
        self.change_directory(newdir,currdir)
        return
    
    def change_dir(self):
        obj =  self.item_get_current()
        if obj == None or not obj.is_dir():
            return
        newdir =  obj.string_get()
        self.change_directory(newdir)
        return

    def enter_button(self):
        obj =  self.item_get_current()
        if obj == None:
            Error("no item available")
            return
        if obj.is_song():
            GlobalDisplayManager.play_song(obj, None, None)
        elif obj.is_dir():
            self.change_dir()
        elif obj.is_playlist():
            GlobalDisplayManager.edit_playlist(self,obj.string_get(),0)
        return

    def hotkey(self,n):
        filename = ConfigPlaylistRoot + ("/%d.m3u" % n)
        obj = song.Item(filename,GlobalMusicus)
        if obj.is_playlist():
            GlobalDisplayManager.edit_playlist(self,obj.string_get(),0)
        else:
            Error("unsupported file " + filename )
        return

    def toggle_playable_only(self):
        if self._playable_only:
            self._playable_only = 0
        else:
            self._playable_only = 1
        self.change_directory( self._cwd )
        return

    def delete_file(self):
        Error("file deletion currently disabled")
        return
    
        file =  self.item_get_current().string_get()
        if os.path.isdir(file): return
        assert(0)

    def add_item(self,item,at_cursor):
        Error("cannot add item to file browser")
        return
    
    def transfer_item(self,recursive,at_cursor):
        obj =  self.item_get_current()
        if obj == None:
            Error("no item available")
            return
        str = obj.string_get()
        if recursive and obj.is_dir() or obj.is_song():
            GlobalDisplayManager.add_to_other_window(self,obj,at_cursor)
            Info("Adding to playlist: " + str )
        else:
            Error("Cannot copy item: " + str )
        self.pos_cursor_move(1,1)
        return
    
    def new_playlist(self):
        Info("new playlist ")
        GlobalDisplayManager.edit_playlist(self,"noname.m3u",1)
        return
    
    def redraw_border(self):
        self.draw_frame()        
        self.draw_header(2,"[DIRECTORY: " + self._cwd +"]")
        self.draw_footer(2,"[" + self._viewmode.string() + "]")

        if self._playable_only:
            self.draw_footer(20,"[playable]")

        self.draw_footer(35,"[" + ("%d" % self._playlist.length()) + "]")
        return

#######################################################
## 
#######################################################

class DisplayManager:

    def __init__(self, directory1,directory2):
        assert( curses.COLS > 70)
        assert( curses.LINES > 20)


        COLUMNS_TOTAL =  curses.COLS
        COLUMNS_LEFT  = COLUMNS_TOTAL/2
        COLUMNS_RIGHT  = COLUMNS_TOTAL - COLUMNS_LEFT

        LINES_TOTAL = curses.LINES
        LINES_SONG = 6
        LINES_STATUS = 6
        LINES_MIDDLE = LINES_TOTAL - LINES_STATUS - LINES_SONG

        ## create root and four main panels
    
        self._root = PanelRoot(None,"Root Window",
                                     0,0,COLUMNS_TOTAL,LINES_TOTAL)


        self._player = PanelSong(self._root,"Current Song Window",
                                 0, 0 ,COLUMNS_TOTAL, LINES_SONG)

        self._right_editor = PanelListPlaylistEditor(self._root, "Right Playlist Editor",
                                                    COLUMNS_LEFT, LINES_SONG, COLUMNS_RIGHT, LINES_MIDDLE)

        self._right_browser = PanelListFilebrowser(self._root, "Right File Browser",
                                                  COLUMNS_LEFT, LINES_SONG, COLUMNS_RIGHT, LINES_MIDDLE)

        self._left_editor = PanelListPlaylistEditor(self._root, "Left Playlist Editor",
                                                     0,LINES_SONG,COLUMNS_LEFT,LINES_MIDDLE)

        self._left_browser = PanelListFilebrowser(self._root, "Left File Browser",
                                                   0,LINES_SONG,COLUMNS_LEFT,LINES_MIDDLE)

        self._status = PanelStatus(self._root,"Status Window",
                                   0, LINES_TOTAL - LINES_STATUS , COLUMNS_TOTAL, LINES_STATUS)

        self._left_browser.change_directory(directory1,"$$",0)
        
        self._right_browser.change_directory(directory2,"$$",0)

        self._left = self._left_browser

        self._right = self._right_browser

        self._focus = self._left
        self._focus.focus_on()

        self._left.visible_on()
        self._right.visible_on()
        self._player.visible_on()
        self._status.visible_on()
        return

    def process_key(self, c):
        return self._root.process_key(c)

    def cycle_focus(self):
        self._focus.focus_off()
        self._focus.redraw()        

        if self._focus == self._left: self._focus = self._right
        elif self._focus == self._right: self._focus = self._status
        elif self._focus == self._status: self._focus = self._player
        elif self._focus == self._player: self._focus = self._left
        else:
            assert(0)

        self._focus.focus_on()
        self._focus.redraw()        
        return

    def status_panel(self):
        return self._status
    
    def move_focus_left(self):
        self._focus.focus_off()
        self._focus.redraw()        

        self._focus = self._left

        self._focus.focus_on()
        self._focus.redraw()        
        return

    def move_focus_right(self):
        self._focus.focus_off()
        self._focus.redraw()        

        self._focus = self._right

        self._focus.focus_on()
        self._focus.redraw()        
        return

    def resize_horizontal(self,delta):
#        Info("resize %d" % delta)
        dim1 = self._left._dim
        w1 = dim1[2] + delta

        dim2 = self._right._dim
        x2 = dim2[0] + delta
        w2 = dim2[2] - delta

        if w1 < 15 or w2 < 15: return
        
        self._left.resize(dim1[0],dim1[1],w1,dim1[3])
        self._left.redraw()

        self._right.resize(x2,dim2[1],w2,dim2[3])
        self._right.redraw()
        return

    def quit(self):
        if self._left_editor.dirty() or  self._right_editor.dirty():
            Error("playlists must be saved or discarded first")
        else:
            sys.exit()
        return
    
    #######################################################
    # misc trans panel functions!
    #######################################################
    def kill_song(self):
        self._player.kill_song()
        return
    
    def play_song(self,obj, a,b):
        return self._player.play_song(obj, a,b)

    def add_to_other_window(self, panel,str,at_cursor):
        if panel == self._left:
            self._right.add_item(str,at_cursor)
        elif panel == self._right:
            self._left.add_item(str,at_cursor)
        else:
            assert(0)
        return

    def exit_playlist_edit(self,panel):
        Info("Returing to file browser")
        assert( panel.has_focus() )
        dim =  panel._dim
        panel.focus_off()
        panel.visible_off()
        if panel == self._left:
            panel = self._left_browser
            self._left = panel
        elif panel == self._right:
            panel = self._right_browser
            self._right = panel
        else:
            assert(0)

        self._focus = panel
        panel.resize( dim[0],dim[1],dim[2],dim[3])
        panel.focus_on()
        panel.visible_on()
        return
    
    def edit_playlist(self,panel,file,is_new):
        assert( panel.has_focus() )
        dim =  panel._dim
        panel.focus_off()
        panel.visible_off()
        
        if panel == self._left:
            panel = self._left_editor
            self._left = panel
        elif panel == self._right:
            panel = self._right_editor
            self._right = panel
        else:
            assert(0)

        self._focus = panel
        panel.resize( dim[0],dim[1],dim[2],dim[3])
        panel.visible_on()
        panel.focus_on()
        panel.load_file(file,is_new)
        return

    def help_strings(self):
        return self._root.help_strings() + \
               self._player.help_strings() + \
               self._left.help_strings() + \
               self._right.help_strings() + \
               self._status.help_strings()  + \
               ["","","", "Keys marked with a '*' are only active when the corresponding window has focus"] 
#               [self._left_editor.string_status(),self._left_browser.string_status(),
#                self._right_editor.string_status(),self._right_browser.string_status()]
    

#######################################################
## 
#######################################################

def usage():
    usage = _("Usage: %s [ filename | playlist.m3u ] ...\n\n\tpress h for a help screen\n")
    sys.stderr.write(usage % sys.argv[0])
    return


#######################################################
## 
#######################################################

GlobalMusicus = None

def cursesmain(stdscr):

    global GlobalDisplayManager
    global GlobalTimer
    global GlobalMusicus
    global ConfigMusicRoot
    global ConfigPlaylistRoot

    Info("##################################################")
    Info("starting cjukebox " + __version__)
    Info("##################################################")

    ## the next few configuartion lines may not work reliably with older python versions
    
    # make getch nonblocking
    stdscr.nodelay(1)

    # make cursor invisible
    # causes problems on some systems
    curses.curs_set(0)
    curses.cbreak()
#    curses.keypad(1)
    curses.noecho()

    if curses.has_colors():

        ## for color constants see /usr/include/ncurses.h

        curses.init_pair(1, curses.COLOR_WHITE, curses.COLOR_BLUE) # no focus color
        curses.init_pair(2, curses.COLOR_WHITE, curses.COLOR_RED)  # focus color
        curses.init_pair(3, curses.COLOR_BLUE, curses.COLOR_WHITE)
        curses.init_pair(4, curses.COLOR_BLACK, curses.COLOR_RED)
        

    ## Prepare Timer

    GlobalTimer = TimerEvents()

    #######################################################
    ## Initialize Musicus including version check
    #######################################################    

    GlobalMusicus = musicus.Musicus()
    musicus_version = GlobalMusicus.version()
    if musicus_version[0] == 0 :
        (major,minor) = map( int, string.split(musicus_version[1][1:-1],".") )
        if major == 1 and minor < 13:
            musicus_version = "Musicus version %d.%d too old, song playback and tag reading will not work" % (major,minor)
            GlobalMusicus = None
        else:
            musicus_version = "Found Musicus version %d.%d" % (major,minor)

    else:
        musicus_version = "Cannot find Musicus, song playback and tag reading will not work"
        GlobalMusicus = None

    if GlobalMusicus: GlobalMusicus.settitleformat("%p - %t")
        
    ## GEOMETRY: managing the screen real estate
    if not os.path.isdir(ConfigMusicRoot):
        Warning("cannot find start directory %s\n" % directory)
        ConfigMusicRoot = "/"

    if not os.path.isdir(ConfigPlaylistRoot):
        Warning("cannot find start directory %s\n" % directory)
        ConfigPlaylistRoot = "/"
        
    GlobalDisplayManager = DisplayManager(ConfigMusicRoot,ConfigPlaylistRoot)


    #######################################################
    ## Print some welcome message in the status window
    #######################################################

    Info("Welcome to " + __version__)

    if GlobalMusicus: Info( musicus_version ) 
    else: Error( musicus_version )

    Info("Press 'h' for help")
    #######################################################
    ## MAIN LOOP
    #######################################################
    #    print stdscr.getmouse() 

    try:
        while 1:
            # for performance reasons
            time.sleep(0.01)
            panel.update_panels()
            curses.doupdate()
            GlobalTimer.poll()

            # key board event processing
            
            c = stdscr.getch()

            if c < 0:
                continue
            elif c == curses.KEY_RESIZE:
                Info("got resize event")
            else:
                if not GlobalDisplayManager.process_key(c):
                    Error("unknown key: %d" % c)
    finally:
        Info("Exiting cjukebox")
        GlobalDisplayManager.kill_song()
        if GlobalMusicus: GlobalMusicus.kill()

    return

#######################################################
## 
#######################################################

if __name__ == "__main__": 
    try:
        opts, args = getopt.getopt(sys.argv[1:], "")
    except:
        usage()
        sys.exit(1)

    whrandom.seed()

    if 1:
        curses.wrapper( cursesmain )
    else:
        import profile
        import pstats

        profile.run("curses.wrapper( cursesmain )","run.prof")    
        p = pstats.Stats("run.prof")
        p.sort_stats('cumulative').print_stats(20)
        p.sort_stats('time').print_stats(20)

    sys.exit(0)
#######################################################
## 
#######################################################

