# -*- python -*-
#
#      Copyright (C) 2002-2004 Robert Muth <robert at muth dot org>
#
#      This program is free software; you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation; version 2 of June 1991.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program in form of the file COPYING; 
#      if not, write to the 
#
#      Free Software Foundation, Inc. <http://www.fsf.org>
#      59 Temple Place, Suite 330, 
#      Boston, MA 02111-1307  USA
#
######################################################
import os
import re

import cddb

#######################################################
##
#######################################################

def SaveToFile(file, message):
    file = open(file,"wa")
    file.write( message )
    file.close()
    return

#######################################################
## 
#######################################################

ConfigMagicFiles = ["id","index","flags","toc","toc.1","toc.2","toc.3"]

#######################################################
## 
#######################################################

class Viewmode:

    modes =[ "tag","cddb","file","path"]
             
    def __init__(self):
        self._mode = 0

    def string(self):
        return Viewmode.modes[ self._mode]

    
    def cycle(self):
        self._mode = self._mode + 1
        if self._mode == len(Viewmode.modes):
            self._mode = 0

#######################################################
## Item Object for use with PanelList Object, etc
#######################################################

class Item:

    TocContainer = cddb.TocContainer()

    use_tag = 1
    use_cddb = 1

    PatternEmptyLine = re.compile(r"^[ \t\r]*$")
    PatternCommentLine = re.compile(r"^#.*$")
    PatternPlayList = re.compile(r"^.*\.m3u$", re.I)


    TYPE_INVALID = 0
    TYPE_DIR  = 1
    TYPE_LIST = 2
    TYPE_SONG = 3

    TYPE_FILE = 4
    TYPE_COMMENT = 5
    TYPE_BADFILE = 6
    TYPE_EMPTY = 7
        
    def __init__(self, string, musicus):
        self._toc = None
        self._string = string
        self._size = 0
        self._length = None   # we use this as an indicator for having either cddb or tag info
        self._tag_title = None
        self._cddb_title = None
        
        match = Item.PatternEmptyLine.search(string)
        if match:
            self._type = Item.TYPE_EMPTY
            return

        match = Item.PatternCommentLine.search(string)
        if match:
            self._type = Item.TYPE_COMMENT
            return

        if os.path.isdir(string):
            if Item.use_cddb:
                self._toc = Item.TocContainer.find_toc(string)
                if self._toc: self._cddb_title = self._toc.title_get() 
            self._type = Item.TYPE_DIR
            return

        match = Item.PatternPlayList.search(string)
        if match:
            self._type = Item.TYPE_LIST
            return

        if os.path.exists(string):
            self._size = os.path.getsize(string)
            (dirname,filename) = os.path.split(string)
            if filename in ConfigMagicFiles:
                self._type = Item.TYPE_FILE
                return 
            else:
                if self.use_cddb:
                    # here we hardcode the cddb format
                    self._toc = Item.TocContainer.find_toc(dirname)
                    if self._toc:
                        (track,ext,time) = self._toc.track_get(self._string)
                        if not track:
                            self._toc = None
                        else:
                            self._type = Item.TYPE_SONG            
                            assert( track and self._string)
                            self._cddb_title = self._toc.artist_get() + " " + track
                            self._length = time
                # we also want to get the tag info
                # possibly overwriting the length
                # since we believe it is more correct
                # even though I know of one case where the
                # opposite is true
                if musicus and self.use_tag:
                    x = musicus.getinfo(string)
                    if x[0] == 0:
                        self._length = int(x[2])/1000
                        self._tag_title = x[3][1:-1]
                        self._type = Item.TYPE_SONG            

                if self._length != None: return                        


        self._type = Item.TYPE_BADFILE
        return

    def string_get(self):
        return self._string

    def string_short_get(self):
        if self._type == Item.TYPE_COMMENT:
            return self._string
        else:
            return  os.path.split(self._string)[1]


    def decoration(self):
        if self._type == Item.TYPE_EMPTY:
            return " " 

        elif self._type == Item.TYPE_SONG:
            if self._length == None:
                return "s"
            else:
                return "S"
            
        elif self._type == Item.TYPE_FILE:
            return "f" 

        elif self._type == Item.TYPE_COMMENT:
            return "c"
        
        elif self._type == Item.TYPE_LIST:
            return "p"

        elif self._type == Item.TYPE_BADFILE:
            return "*"

        elif self._type == Item.TYPE_DIR:
            if self._toc == None:
                return "d" 
            else:
                return "D" 
        else:
            assert(0)
            return
        
    def dir_string_fancy_get(self,viewmode):
        if viewmode == "cddb" and self._toc:
            return "[" + self._cddb_title + "]"
        elif viewmode == "path":
            return self.string_get()
        else:
            return self.string_short_get()

    def list_string_fancy_get(self,viewmode):
        if viewmode == "path":
            return self.string_get()
        else:
            return self.string_short_get()
        
    def default_string_fancy_get(self,viewmode):
        if viewmode == "path":
            return self.string_get()
        else:
            return self.string_short_get()

    def song_string_fancy_get(self,viewmode,showtime):
        assert( self._length != None)
        if showtime: time  = "%2d:%02d " % divmod(self._length,60)
        else: time = ""
        
        if viewmode == "cddb" and self._toc:
            return time + "[" + self._cddb_title + "]"
        elif viewmode == "tag" and self._tag_title:
            return time + "[" + self._tag_title + "]"
        elif viewmode == "path":
            return time + self.string_get()
        else:
            return time + self.string_short_get()
       
    def string_fancy_get(self,viewmode,showtime=1):
        viewmode = viewmode.string()
        ## we should probably have a virtual function instead of this
        ## dispatch
        
        if self._type == Item.TYPE_SONG: s = self.song_string_fancy_get(viewmode,showtime) 
        elif  self._type == Item.TYPE_LIST: s = self.list_string_fancy_get(viewmode)
        elif  self._type == Item.TYPE_DIR: s = self.dir_string_fancy_get(viewmode)
        else: s  = self.default_string_fancy_get(viewmode)

        return self.decoration() + " " + s

    def time(self):
        if self._length:
            return self._length
        else:
            return 0
        
    def description_get(self,viewmode):
        s = self.string_fancy_get(viewmode,0)
        s = re.sub(r'[_\[\]]', " ", s)            
        s = re.sub(r'[/:()]', ".", s)            
        s = re.sub(r'\'', r'\'', s)
        s = re.sub(r'\&', r'and', s)
        return s
            
    def type_get(self):
        return self._type

    def is_song(self):
        return self._type == Item.TYPE_SONG

    def is_dir(self):
        return self._type == Item.TYPE_DIR

    def is_playlist(self):
        return self._type == Item.TYPE_LIST
    
#######################################################
## PLAYLIST
#######################################################

class Playlist:
    
    def __init__(self):
        self._items = []
        self._filename = "noname"
        self._dirty = 0
        return

    def clear(self,filename):
        self._items = []
        self._filename = filename
        self._dirty = 0
        return
    
    def contains(self,item):
        for i in self._items:
            if item.string_get() == i.string_get(): return 1
            return 0

    def item(self,n):
        return self._items[n]

    def filename(self):
        return self._filename

    def items(self):
        return self._items

    def append(self,item):
        self._items.append( item )
        self._dirty = 1
        return

    def swap(self,pos1,pos2):
        item1 = self._items[pos1]
        item2 = self._items[pos2]
        self._items[pos1] = item2
        self._items[pos2] = item1
        return
    
    def insert(self,pos,item):
        self._items.insert(pos, item )
        self._dirty = 1
        return

    def delete(self,pos):
        del self._items[pos]
        self._dirty = 1
        return
    
    def length(self):
        return len(self._items)

    def time(self):
        accumulator = 0
        for i in self._items:
            accumulator = accumulator + i.time()
        return accumulator
        
    def dirty(self):
        return self._dirty

    def clear_dirty_flag(self):
        self._dirty = 0
        return
    
    def load(self,filename,musicus):
        self._items = []
        self._filename = filename
        self._dirty = 0            
        
        try:
            input = open(filename, "r")
            lines = input.readlines()
            input.close()
        except:
            return
        
        for l in lines:
            item = Item(l[:-1],musicus)
            self._items.append(item);
        return
   
    def save(self,filename=None):
        #try:
        if not filename:
            filename = self._filename
        input = open(filename, "w+")
        for i in self._items:
            input.write( i.string_get()+"\n"); 
        input.close()
        self._dirty = 0
        return

#######################################################
## eof
#######################################################
